#include "expression_code.h"
#include <stdexcept>

///////////////////////////////////////
//
// call expression
//
//////////////////////////////////////

expression::call::call(base* operation) : m_operation(operation)
{
}

expression::call::~call()
{
	delete m_operation;
}

binary::link expression::call::link(linker::context& c)
{
	std::vector<binary::link> ops;
	ops.push_back(m_operation->link(c));

	for (size_t i = 0; i < m_operands.size(); i++)
	{
		ops.push_back(m_operands[i]->link(c));
	}

	return c.add(linker::object::create_call(estimate(c), ops.size(), &ops[0]));
}

int expression::call::estimate(linker::context& c)
{
	int est = base::estimate(c);
	
	// evaluate function
	est += m_operation->estimate(c);

	// apply function
	est += m_operation->estimateApply(c, m_operands.size(), m_operands.get());

	return est;
}

void expression::call::add(base* e)
{
	m_operands.push_back(e);
}

///////////////////////////////////////
//
// if expression
//
//////////////////////////////////////

expression::if_::if_(base* condition, base* cons, base* alt)
{
	m_condition = condition;
	m_consequence = cons;
	m_alternate = alt;
}

expression::if_::~if_()
{
	delete m_condition;
	delete m_consequence;
	delete m_alternate;
}

binary::link expression::if_::link(linker::context& c)
{
	return c.add(linker::object::create_if(
				estimate(c),
				m_condition->link(c),
				m_consequence->link(c),
				m_alternate->link(c)));
}

int expression::if_::estimate(linker::context& c)
{
	int est = base::estimate(c);

	// conditional will be evaluated always
	est += m_condition->estimate(c);

	// consequence or alternate? we don't know. Choose max
	int est_cons = m_consequence->estimate(c);
	int est_alt = m_alternate->estimate(c);

	est += (est_cons > est_alt)? est_cons : est_alt;

	return est;
}

//////////////////////////////////////////
//
// cond expression
//
//////////////////////////////////////////

expression::cond::cond() : m_else(0)
{
}

expression::cond::~cond()
{
	delete m_else;
}

binary::link expression::cond::link(linker::context& c)
{
	std::vector<binary::link> links1;
	std::vector<binary::link> links2;

	{for (size_t i = 0; i < m_test.size(); i++)
	{
		links1.push_back(m_test[i]->link(c));
		links2.push_back(m_exp[i]->link(c));
	}}

	if (m_else)
	{
		links1.push_back(c.add(linker::object::create(true)));
		links2.push_back(m_else->link(c));
	}

	return c.add(linker::object::create_cond(estimate(c), links1.size(), &links1[0], &links2[0]));
}

int expression::cond::estimate(linker::context& c)
{
	// we don't know, which test will successed,
	// so we summ all estimates for tests and divide by 2
	int test_est = 0;
	
	{for (size_t i = 0; i < m_test.size(); i++)
	{
		test_est += m_test[i]->estimate(c);
	}}

	test_est /= 2;
	
	// we don't know, which test will successed,
	// so we summ all estimates for expressions and divide
	// by quantity of expressions
	int exp_est = 0;

	{for (size_t i = 0; i < m_exp.size(); i++)
	{
		exp_est += m_exp[i]->estimate(c);
	}}

	if (m_else)
	{
		exp_est += m_else->estimate(c);
		exp_est /= (int)m_exp.size() + 1;
	}
	else
	{
		exp_est /= (int)m_exp.size();
	}

	return base::estimate(c) + test_est + exp_est;
}

void expression::cond::addClause(base* c, base* e)
{
	m_test.push_back(c);
	m_exp.push_back(e);
}

void expression::cond::addElse(base* e)
{
	m_else = e;
}

/////////////////////////////////////////////
//
// or expression
//
/////////////////////////////////////////////

binary::link expression::or_::link(linker::context& c)
{
	std::vector<binary::link> links;

	for (size_t i = 0; i < m_test.size(); i++)
	{
		links.push_back(m_test[i]->link(c));
	}

	return c.add(linker::object::create_or(estimate(c), links.size(), &links[0]));
}

int expression::or_::estimate(linker::context& c)
{
	int est = base::estimate(c);
	
	// we don't know, which test will successed,
	// so we summ all estimates for tests and divide by 2
	{for (size_t i = 0; i < m_test.size(); i++)
	{
		est += m_test[i]->estimate(c);
	}}

	est /= 2;
	
	return est;
}

void expression::or_::add(base* t)
{
	m_test.push_back(t);
}

//////////////////////////////////////////////////
//
// and expression
//
//////////////////////////////////////////////////

binary::link expression::and_::link(linker::context& c)
{
	std::vector<binary::link> links;

	for (size_t i = 0; i < m_test.size(); i++)
	{
		links.push_back(m_test[i]->link(c));
	}

	return c.add(linker::object::create_and(estimate(c), links.size(), &links[0]));
}

int expression::and_::estimate(linker::context& c)
{
	int est = base::estimate(c);

	// we don't know, which test will successed,
	// so we summ all estimates for tests and divide by 2
	{for (size_t i = 0; i < m_test.size(); i++)
	{
		est += m_test[i]->estimate(c);
	}}

	est /= 2;
	
	return est;
}

void expression::and_::add(base* t)
{
	m_test.push_back(t);
}

