#include "lexer.h"
#include <stdexcept>

namespace lexer
{
	bool isNonSpaceDelimeter(char c)
	{
		return ((c == '(') || (c == ')') || (c == ';') || (c == '\''));
	}

	bool toNumber(const std::string& str, double* d);
	bool isIdentifier(const std::string& str);

	token getString(stream::input<char>& input);
	token getMultiSymbolToken(int first, stream::input<char>& input);
	void skipComment(stream::input<char>& input);

	std::string getSolidString(int first, stream::input<char>& input);	
}

//////////////////////////////////////////////
//
// Public functions
//
//////////////////////////////////////////////

lexer::token_source::token_source(stream::source<char>& i) : m_input(i)
{
}

lexer::token lexer::token_source::get()
{
	if (m_eofReached)
	{
		//    
		throw stream::eof();
	}

	for (;;)
	{
		char c;

		try
		{
			c = m_input.get();
		}
		catch (stream::eof)
		{
			m_eofReached = true;
			throw stream::eof();
		}

		if (!isspace(c))
		{
			switch (c)
			{
			case ';':  skipComment(); break;

			case '(':  return token(token::open_bracket);
			case ')':  return token(token::close_bracket);
			case '\'': return token(token::quote);
			case '.':  return token(token::point);
			
			default: return getMultiSymbolToken(c);
			}
		}
	}
}

//////////////////////////////////////////////
//
// Private functions
//
//////////////////////////////////////////////

lexer::token lexer::token_source::getMultiSymbolToken(char first)
{
	if (first == '\"')
	{
		//   
		return getString();
	}

	std::string str = getSolidString(first);

	// Boolean
	if (str == "#t")
	{
		return token(true);
	}
	else if (str == "#f")
	{
		return token(false);
	}
	
	// Keywords
	if (str == "lambda")
	{
		return token(token::lambda, str);
	}
	else if (str == "if")
	{
		return token(token::if_, str);
	}
	else if (str == "cond")
	{
		return token(token::cond, str);
	}
	else if (str == "or")
	{
		return token(token::or_, str);
	}
	else if (str == "and")
	{
		return token(token::and_, str);
	}
	else if (str == "else")
	{
		return token(token::else_, str);
	}
	else if (str == "define")
	{
		return token(token::define, str);
	}

	// Character
	if ((str.length() == 3) && (str[0] == '#') && (str[1] == '\\'))
	{
		return token(str[2]);
	}
	else if (str == "#\\space")
	{
		return token(' ');
	}
	else if (str == "#\\newline")
	{
		return token('\n');
	}
	
	// numeric
	double d = 0;
	
	if (toNumber(str, &d))
	{
		return token(d);
	}

	// identifier
	if (isIdentifier(str))
	{
		return token(token::variable, str);
	}

	throw std::runtime_error("lexer:Unknown token");
}

lexer::token lexer::token_source::getString()
{
	std::string str;

	try
	{
		for (;;)
		{
			int c = m_input.get();

			if (c == '\"')
			{
				// End of string
				break;
			}

			if (c == '\\')
			{
				// Special character
				c = m_input.get();

				if ((c != '\"') && (c != '\\'))
				{
					// Unknown special character
					throw std::runtime_error("lexer:string:Unknown special character");;
				}
			}

			str.push_back(c);
		}
	}
	catch (stream::eof&)
	{
		// Unexcepted end of file
		throw std::runtime_error("lexer:string:Unexcepted end of file");
	}

	return token(token::string, str);
}

void lexer::token_source::skipComment()
{
	try
	{
		for (;;)
		{
			int c = m_input.get();

			if (c == '\n')
			{
				break;
			}
		}
	}
	catch (stream::eof&)
	{
	}
}

std::string lexer::token_source::getSolidString(char first)
{
	std::string str;
	str.push_back(first);

	try
	{
		for (;;)
		{
			int c = m_input.get();

			if (isspace(c))
			{
				break;
			}

			if (isNonSpaceDelimeter(c))
			{
				m_input.putback(c);
				break;
			}

			str.push_back(c);
		}
	}
	catch (stream::eof&)
	{
	}

	return str;
}

//////////////////////////////////////////
//
// Utility functions
//
////////////////////////////////////////////

bool lexer::toNumber(const std::string& str, double* d)
{
	char* ptr = (char*)str.c_str();
	*d = strtod(str.c_str(), &ptr);

	return (ptr - str.c_str() == str.length());
}

bool lexer::isIdentifier(const std::string& str)
{
	return true;
}