#include "linker_object.h"

namespace
{
	const int base_estimate = 1;
}

///////////////////////////////////////////////////////
//
// Static creator functions
//
///////////////////////////////////////////////////////

///////////
//
// Variable
//
///////////

linker::object* linker::object::create(const std::string& name, binary::link l)
{
	return new object(binary::types::link, base_estimate, name, &l, sizeof(l));
}

///////////
//
// Literals
//
///////////

linker::object* linker::object::create(double v)
{
	binary::numeric b;
	b.value = v;
	return new object(binary::types::numeric, base_estimate, &b, sizeof(b));
}

linker::object* linker::object::create(char v)
{
	binary::character b;
	b.value = v;
	return new object(binary::types::character, base_estimate, &b, sizeof(b));
}

linker::object* linker::object::create(bool v)
{
	binary::boolean b;
	b.value = v;
	return new object(binary::types::boolean, base_estimate, &b, sizeof(b));
}

linker::object* linker::object::create_string(const std::string& str)
{
	return new object(binary::types::string, base_estimate, str.c_str(), str.length());
}

linker::object* linker::object::create_symbol(const std::string& str)
{
	return new object(binary::types::symbol, base_estimate, str.c_str(), str.length());
}

linker::object* linker::object::create_pair(int estimate, binary::link car, binary::link cdr)
{
	binary::pair b;
	b.car = car;
	b.cdr = cdr;
	return new object(binary::types::pair, estimate, &b, sizeof(b));
}

linker::object* linker::object::create_empty_list()
{
	return new object(binary::types::empty_list, base_estimate, 0, 0);
}

///////////
//
// Code
//
///////////

linker::object* linker::object::create_lambda(size_t count, bool list, binary::link exp)
{
	binary::lambda b;

	b.argc.count = (unsigned short)count;
	b.argc.list = (list) ? 1 : 0;
	b.exp = exp;

	return new object(binary::types::lambda, base_estimate, &b, sizeof(b));
}

linker::object* linker::object::create_call(int estimate, size_t count, const binary::link* ops)
{
	return create_countAndLinks(binary::types::call, estimate, count, ops);
}

linker::object* linker::object::create_if(int estimate, binary::link cond, binary::link cons, binary::link alt)
{
	binary::if_ b;
	b.condition = cond;
	b.consequence = cons;
	b.alternate = alt;

	return new object(binary::types::if_, estimate, &b, sizeof(b));
}

linker::object* linker::object::create_cond(int estimate, size_t count, const binary::link* tests, const binary::link* exps)
{
	std::vector<binary::link> links(count * 2);

	for (size_t i = 0; i < count; i++)
	{
		links[2 * i]     = tests[i];
		links[2 * i + 1] = exps[i];
	}

	return create_countAndLinks(binary::types::cond, estimate, links.size(), &links[0]);
}

linker::object* linker::object::create_or(int estimate, size_t count, const binary::link* tests)
{
	return create_countAndLinks(binary::types::or_, estimate, count, tests);
}

linker::object* linker::object::create_and(int estimate, size_t count, const binary::link* tests)
{
	return create_countAndLinks(binary::types::and_, estimate, count, tests);
}

/////////////////////////////////////////////
//
// Public functions
//
/////////////////////////////////////////////

short linker::object::flags() const
{
	if (m_named) 
	{
		return binary::record::named;
	}

	return 0;
}

short linker::object::type() const
{
	return m_type;
}

int linker::object::estimate() const
{
	return m_estimate;
}

int linker::object::size() const
{
	return (int)m_data.size();
}

const char* linker::object::get() const
{
	if (size())
	{
		return &m_data[0];
	}

	return 0;
}

const std::string& linker::object::name() const
{
	return m_name;
}

bool linker::object::named() const
{
	return m_named;
}

//////////////////////////////////////////////////
//
// Private functions
//
//////////////////////////////////////////////////

linker::object::object(int t, unsigned int e, const void* c, size_t s) : m_type(t), m_named(false)
{
	m_estimate = e;
	const char* tmp = (const char*)c;
	m_data.assign(tmp, tmp + s);
}

linker::object::object(int t, unsigned int e, const std::string& name, const void* c, size_t s) 
			: m_type(t), m_named(true), m_name(name)
{
	m_estimate = e;
	const char* tmp = (const char*)c;
	m_data.assign(tmp, tmp + s);
}

linker::object* linker::object::create_countAndLinks(int type, int estimate, size_t count, const binary::link* ops)
{
	object* obj = new object(type, estimate, 0, 0);

	binary::call* pb;

	obj->m_data.resize(sizeof(*pb) + sizeof(binary::link) * (count - 1));
	
	pb = reinterpret_cast<binary::call*>(&(obj->m_data[0]));

	pb->argc = (unsigned short)count;
	memcpy(pb->ops, ops, count * sizeof(binary::link));
	
	return obj;
}
