#include "syntacter_code.h"
#include "syntacter_lambda.h"
#include "../../__common/binaries.h"
#include "../expression/expression_code.h"

#include <stdexcept>

//////////////////////////////////////////////////
//
// Public functions
//
///////////////////////////////////////////////////

//
// Read expression from stream
expression::base* syntacter::read_code(stream::input<lexer::token>& i)
{
	// Call | lambda | if | cond | and | or
	lexer::token t = i.get();

	if ((t.type() == lexer::token::open_bracket) || (t.type() == lexer::token::variable))
	{
		// Call
		i.putback(t);
		return expression::call::read(i);
	}

	switch (t.type())
	{
	case lexer::token::lambda: // lambda
		return read_lambda(i);
	case lexer::token::if_: // if
		return expression::if_::read(i);
	case lexer::token::cond: // cond
		return expression::cond::read(i);
	case lexer::token::or_:  // or
		return expression::or_::read(i);
	case lexer::token::and_: // and
		return expression::and_::read(i);
	}

	throw std::runtime_error("syntacter:read_code: Unknown expression");
}

////////////////////////////////////////////////////
//
// Concrete expression
//
/////////////////////////////////////////////////////

////////////
//
// Call expression
//
////////////

//
// 	<procedure call> ::= (<operator> <operand>*)
//		<operator> ::= <expression>
//		<operand> ::= <expression>
//

expression::call* expression::call::read(stream::input<lexer::token>& input)
{
	// Read operation
	std::auto_ptr<call> ce(new call(syntacter::read(input)));

	// Read all operands
	for (;;)
	{
		lexer::token t = input.get();

		if (t.type() == lexer::token::close_bracket)
		{
			break;
		}
		
		input.putback(t);
		ce->add(syntacter::read(input));
	}

	return ce.release();
}

//////////////
//
// If expression
//
//////////////

//
//	<conditional> ::= (if <test> <consequent> <alternate>)
//		<test> ::= <expression>
//		<consequence> ::= <expression>
//		<alternate> ::= <expression>
//

expression::if_* expression::if_::read(stream::input<lexer::token>& input)
{
	std::auto_ptr<base> test(syntacter::read(input));
	std::auto_ptr<base> consequence(syntacter::read(input));
	std::auto_ptr<base> alternate(syntacter::read(input));

	lexer::token t = input.get();

	if (t.type() != lexer::token::close_bracket)
	{
		throw std::runtime_error("expression::if_::read: waiting for )");
	}

	return new if_(test.release(), consequence.release(), alternate.release());
}

//////////////
//
// Cond expression
//
//////////////

//
//	<cond> ::= (cond <cond clause>+) | (cond <cond clause>* (else <expression>))
//		<cond clause> ::= (<test> <expression>)
//

expression::cond* expression::cond::read(stream::input<lexer::token>& input)
{
	std::auto_ptr<cond> ce(new cond());

	for (int i = 0;; i++)
	{
		lexer::token t = input.get();

		if (t.type() == lexer::token::close_bracket)
		{
			// end of cond
			if (i == 0)
			{
				// empty cond
				throw std::runtime_error("expression::cond::read: empty cond");
			}
			break;
		}

		if (t.type() != lexer::token::open_bracket)
		{
			// bug
			throw std::runtime_error("expression::cond::read: waiting for (");
		}

		t = input.get();

		if (t.type() == lexer::token::else_)
		{
			// else clause
			ce->addElse(syntacter::read(input));
			
			t = input.get();

			// end of else
			if (t.type() != lexer::token::close_bracket)
			{
				throw std::runtime_error("expression::cond::read: waiting for )");
			}

			t = input.get();

			// end of cond
			if (t.type() != lexer::token::close_bracket)
			{
				throw std::runtime_error("expression::cond::read: waiting for )");
			}

			break;
		}
		else
		{
			input.putback(t);

			std::auto_ptr<base> c(syntacter::read(input));
			std::auto_ptr<base> e(syntacter::read(input));

			t = input.get();

			// end of clause
			if (t.type() != lexer::token::close_bracket)
			{
				throw std::runtime_error("expression::cond::read: waiting for )");
			}

			ce->addClause(c.release(), e.release());
		}
	}

	return ce.release();
}

//////////////
//
// Or expression
//
//////////////

//
// <or> ::= (or <test>*)
//

expression::or_* expression::or_::read(stream::input<lexer::token>& input)
{
	std::auto_ptr<or_> oe(new or_());

	for (;;)
	{
		lexer::token t = input.get();

		if (t.type() == lexer::token::close_bracket)
		{
			// end of or
			break;
		}

		input.putback(t);
		oe->add(syntacter::read(input));
	}

	return oe.release();
}

//////////////
//
// And expression
//
//////////////

//
// <and> ::= (and <test>*)
//

expression::and_* expression::and_::read(stream::input<lexer::token>& input)
{
	std::auto_ptr<and_> ae(new and_());

	for (;;)
	{
		lexer::token t = input.get();

		if (t.type() == lexer::token::close_bracket)
		{
			// end of and
			break;
		}

		input.putback(t);
		ae->add(syntacter::read(input));
	}

	return ae.release();
}