#include "object_base.h"
#include "../../__common/objects_types.h"
#include "../../__common/binaries.h"
#include "../../__common/standard_functions_codes.h"

#include <stdexcept>

namespace object
{
	std::string link2str(binary::link l)
	{
		if (l.l.table == binary::link::local)
		{
			char buff[20];
			sprintf(buff, "l%d", l.l.item);
			return buff;
		}

		if (l.l.table == binary::link::program)
		{
			char buff[20];
			sprintf(buff, "p%04d", l.l.item);
			return buff;
		}

		if (l.l.table == binary::link::standard)
		{
			switch (l.l.item)
			{
				case binary::stdfunc::equivalence: return "=";

				case binary::stdfunc::plus: return "+";
				case binary::stdfunc::minus: return "-";
				case binary::stdfunc::multiplication: return "*";
				case binary::stdfunc::division: return "/";
				case binary::stdfunc::sine: return "sin";
				case binary::stdfunc::cosine: return "cos";
				case binary::stdfunc::tangent: return "tan";
				case binary::stdfunc::arcsine: return "asin";
				case binary::stdfunc::arccosine: return "acos";
				case binary::stdfunc::arctangent: return "atan";
				case binary::stdfunc::exponent: return "exp";
				case binary::stdfunc::logarithm: return "log";
				
				case binary::stdfunc::car: return "car";
				case binary::stdfunc::cdr: return "cdr";
				case binary::stdfunc::cons: return "cons";
				case binary::stdfunc::list: return "list";
				case binary::stdfunc::length: return "length";
				case binary::stdfunc::map: return "map";

				case binary::stdfunc::isNull: return "null?";
				case binary::stdfunc::isReal: return "real?";
				case binary::stdfunc::isPair: return "pair?";
			}

			char buff[20];
			sprintf(buff, "s%02d", l.l.item);
			return buff;
		}

		throw std::runtime_error("unknown link table");
	}

	//
	// Numeric object
	class numeric 
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size != sizeof(binary::numeric))
			{
				throw std::runtime_error("object::numeric::create: wrong size!");
			}

			const binary::numeric* p = (const binary::numeric*)v;

			char buff[100];
			sprintf(buff, "numeric: %lf", p->value);
			return buff;
		}
	};

	//
	// Boolean object
	class boolean
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size != sizeof(binary::boolean))
			{
				throw std::runtime_error("object::boolean::create: wrong size!");
			}

			const binary::boolean* p = (const binary::boolean*)v;

			if (p->value)
			{
				return "bool: #t";
			}

			return "bool: #f";
		}
	};

	//
	// Character object
	class character
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size != sizeof(binary::character))
			{
				throw std::runtime_error("object::character::create: wrong size!");
			}

			const binary::character* p = (const binary::character*)v;

			if (p->value == ' ')
			{
				return "character: #space";
			}
			else if (p->value == '\n')
			{
				return "character: #newline";
			}

			char buff[20];
			sprintf(buff, "character: %c", p->value);
			return buff;
		}
	};

	//
	// String object
	class string
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			std::string str;
			str.assign(v, v + size);

			std::string res = "string: \"";
			res += str;
			res += "\"";

			return res;
		}
	};

	//
	// Symbol object
	class symbol
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size == 0)
			{
				throw std::runtime_error("object::symbol::create: wrong size!");
			}

			std::string str;
			str.assign(v, v + size);

			std::string res = "symbol: ";
			res += str;

			return res;
		}
	};

	//
	// Pair object
	class pair
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size != sizeof(binary::pair))
			{
				throw std::runtime_error("object::pair::create: wrong size!");
			}

			const binary::pair* p = (const binary::pair*)v;

			std::string res = "pair: (";
			res += link2str(p->car);
			res += " . ";
			res += link2str(p->cdr);
			res += ")";

			return res;
		}
	};

	//
	// Empty_list object
	class empty_list
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size != 0)
			{
				throw std::runtime_error("object::empty_list::create: wrong size!");
			}

			return "empty list: ()";
		}
	};

	// Call object
	class call
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size < sizeof(binary::call))
			{
				throw std::runtime_error("object::call::create: wrong size!");
			}

			const binary::call* p = (const binary::call*)v;

			if (size != sizeof(binary::call) + (p->argc - 1) * sizeof(binary::link))
			{
				throw std::runtime_error("object::call::create: wrong size!");
			}

			std::string res = "call: (";
			res += link2str(p->ops[0]);

			for (int i = 1; i < p->argc; i++)
			{
				res += " ";
				res += link2str(p->ops[i]);
			}

			res += ")";
			return res;
		}
	};

	//
	// if_ object
	class if_
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size < sizeof(binary::if_))
			{
				throw std::runtime_error("object::if_::create: wrong size!");
			}

			const binary::if_* p = (const binary::if_*)v;

			std::string res = "if: (if ";
			res += link2str(p->condition);
			res += " ";
			res += link2str(p->consequence);
			res += " ";
			res += link2str(p->alternate);
			res += ")";
			
			return res;
		}
	};

	// Lambda object
	class lambda
	{
	public:

		static std::string create(const char* v, size_t size)
		{
			if (size != sizeof(binary::lambda))
			{
				throw std::runtime_error("object::lambda::create: wrong size!");
			}

			const binary::lambda* p = (const binary::lambda*)v;

			std::string res = "lambda: (lambda ";

			if (p->argc.count == 0)
			{
				res += "()";
			}
			else if ((p->argc.count == 1) && (p->argc.list != 0))
			{
				res += "x";
			}
			else 
			{
				res += "(x0";

				if (p->argc.list != 0)
				{
					char buff[20];

					{for (int i = 1; i < p->argc.count - 1; i++)
					{
						sprintf(buff, " x%d", i);
						res += buff;
					}}

					sprintf(buff, " . x%d", p->argc.count - 1);
					res += buff;
				}
				else
				{
					char buff[20];

					{for (int i = 1; i < p->argc.count; i++)
					{
						sprintf(buff, " x%d", i);
						res += buff;
					}}
				}

				res += ")";
			}

			res += " ";
			res += link2str(p->exp);
			res += ")";
			
			return res;
		}
	};
}

//
// Common create
//

std::string object::create(const loader::record& rec)
{
	switch (rec.type())
	{
		// literals
		case binary::types::numeric:    return numeric::create(rec.data(), rec.size());
		case binary::types::string:     return string::create(rec.data(), rec.size());
		case binary::types::symbol:     return symbol::create(rec.data(), rec.size());
		case binary::types::character:  return character::create(rec.data(), rec.size());
		case binary::types::boolean:    return boolean::create(rec.data(), rec.size());
		case binary::types::pair:       return pair::create(rec.data(), rec.size());
		case binary::types::empty_list: return empty_list::create(rec.data(), rec.size());

		// code
		case binary::types::call: return call::create(rec.data(), rec.size());
		case binary::types::if_:  return if_::create(rec.data(), rec.size());
		case binary::types::cond: break;
		case binary::types::or_:  break;
		case binary::types::and_: break;

		// lambda
		case binary::types::lambda: return lambda::create(rec.data(), rec.size());
	}

	if (rec.type() == binary::types::link)
	{
		// link
		if (rec.size() != sizeof(binary::link))
		{
			throw std::runtime_error("object::link::create: wrong size!");
		}

		const binary::link* pl = (const binary::link*)rec.data();
		std::string res = "link: ";
		res += link2str(*pl);

		return res;
	}

	throw std::runtime_error("object::create: unknown object type");
}
