#ifndef executer_tasks_critical_section_h
#define executer_tasks_critical_section_h

#ifdef _WIN32

#ifndef _WINDOWS_
#include <windows.h>
#endif //_WINDOWS_

#else  //_WIN32
#include <pthread.h>
#endif //_WIN32

//
// Critical section
// Windows-Unix portable 
class critical_section
{
public:

	critical_section()
	{
#ifdef _WIN32
		InitializeCriticalSection(&m_cs);
#else
		pthread_mutex_init(&m_cs, 0);
#endif
	}

	~critical_section()
	{
#ifdef _WIN32
		DeleteCriticalSection(&m_cs);
#else
		pthread_mutex_destroy(&m_cs);
#endif
	}

	void lock()
	{
#ifdef _WIN32
		EnterCriticalSection(&m_cs);
#else
		pthread_mutex_lock(&m_cs);
#endif	
	}

	bool tryLock()
	{
#ifdef _WIN32
		return (TryEnterCriticalSection(&m_cs) != 0);
#else
		// ???
		return (pthread_mutex_trylock(&m_cs) == 0);
#endif	
	}

	void unlock()
	{
#ifdef _WIN32
		LeaveCriticalSection(&m_cs);
#else
		pthread_mutex_unlock(&m_cs);
#endif
	}

private:

#ifdef _WIN32
	CRITICAL_SECTION m_cs;
#else
	pthread_mutex_t m_cs;	
#endif
};

//
// Auto-locker for critical section
class locker
{
public:

	locker(critical_section& cs) : m_cs(cs)
	{
		m_cs.lock();
	}

	~locker()
	{
		m_cs.unlock();
	}

private:

	critical_section& m_cs;
};

//
// Auto-unlocker for critical section
class unlocker
{
public:

	unlocker(critical_section& cs) : m_cs(cs)
	{
	}

	~unlocker()
	{
		m_cs.unlock();
	}

private:

	critical_section& m_cs;
};

#endif
