#include "pool.h"
#include "thread.h"
#include "critical_section.h"

#include <algorithm>
#include <stdexcept>
#include <stdlib.h>

namespace tasks
{
	//
	// Thread executer
	//
	class thread_executer : public executer, private thread
	{
	public:

		thread_executer(pool& p) : executer(p)
		{
			m_stop = false;
			start();
		}

		~thread_executer()
		{
			m_stop = true;
			waitForEnd();
		}

	protected:

		virtual long func()
		{
			for (;;)
			{
				if (m_stop) break;

				if (!m_p.makeNextTask())
				{
					yield();
				}
			}

			return 0;
		}

	public:

		bool m_stop;
	};
}

//
// Get global pool
//

tasks::pool& tasks::getPool()
{
	static pool p;
	return p;
}

////////////////////////////////////////////////
//
// Pool
//
////////////////////////////////////////////////

tasks::pool::~pool()
{
	// stop all threads before destroy other
	for (size_t i = 0; i < m_execs.size(); i++)
	{
		delete m_execs[i];
	}
}

void tasks::pool::init(size_t count)
{
	m_execs.resize(count);

	for (size_t i = 0; i < m_execs.size(); i++)
	{
		m_execs[i] = new thread_executer(*this);
	}
}

size_t tasks::pool::size() const
{
	return m_execs.size();
}

// Add new delayed task to pool, return task ID
int tasks::pool::push_task(object::proxy p)
{
	locker l1(m_tasksCs);
	locker l2(m_pendingCs);

	m_id++;
	m_tasks[m_id] = smartptr<task>(new task(p));
	m_pending.push_back(m_id);

	return m_id;
}

// get result
object::proxy tasks::pool::getResult(int id)
{
	smartptr<task> pt;
	
	{
		locker l(m_tasksCs);

		// find id in pool
		tasks_pool::iterator i = m_tasks.find(id);

		if (i == m_tasks.end())
		{
			throw std::runtime_error("tasks::pool::getResult: can not find id in m_tasks!");
		}

		pt = i->second;

		if (pt->isReady())
		{
			// task is ready
			return pt->get();
		}
	}

	// task is not ready - do it, or wait
	pt->make();

	return pt->get();
}

// get next pending task
bool tasks::pool::makeNextTask()
{
	int id = 0;
	
	{
		locker l(m_pendingCs);

		if (m_pending.size() == 0)
		{
			return false;
		}

		// get random index
		int index = (rand() % (int)m_pending.size());

		// get ID by this index
		id = m_pending[index];

		// clear queue element
		m_pending[index] = m_pending[m_pending.size() - 1];
		m_pending.pop_back();
	}
	
	smartptr<task> pt;

	{
		locker l(m_tasksCs);

		// find id in pool
		tasks_pool::iterator i = m_tasks.find(id);

		if (i == m_tasks.end())
		{
			throw std::runtime_error("tasks::pool::getNextTask: can not find id in m_tasks!");
		}

		pt = i->second;
	}

	pt->tryMake();
	return true;
}